==> experience_builder-0.2.1-alpha1.tar.gz: experience_builder/src/Entity/Pattern.php <==
<?php

declare(strict_types=1);

namespace Drupal\experience_builder\Entity;

use Drupal\Component\Utility\Random;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\experience_builder\ClientSideRepresentation;
use Drupal\experience_builder\Controller\ApiConfigControllers;
use Drupal\experience_builder\Controller\ClientServerConversionTrait;
use Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItem;
use Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItemInstantiatorTrait;

/**
 * @ConfigEntityType(
 *    id = \Drupal\experience_builder\Entity\Pattern::PLUGIN_ID,
 *    label = @Translation("Pattern"),
 *    label_singular = @Translation("pattern"),
 *    label_plural = @Translation("patterns"),
 *    label_collection = @Translation("Patterns"),
 *    admin_permission = \Drupal\experience_builder\Entity\Pattern::ADMIN_PERMISSION,
 *    handlers = {
 *      "access" = \Drupal\experience_builder\EntityHandlers\ContentCreatorVisibleXbConfigEntityAccessControlHandler::class
 *    },
 *    entity_keys = {
 *      "id" = "id",
 *      "label" = "label",
 *    },
 *    config_export = {
 *      "id",
 *      "label",
 *      "component_tree",
 *    },
 *  )
 */
final class Pattern extends ConfigEntityBase implements XbHttpApiEligibleConfigEntityInterface, ComponentTreeEntityInterface {

  public const string PLUGIN_ID = 'pattern';
  public const string ADMIN_PERMISSION = 'administer patterns';

  use ComponentTreeItemInstantiatorTrait;
  use ClientServerConversionTrait;

  /**
   * Pattern entity ID.
   */
  protected string $id;

  /**
   * The human-readable label of the Experience Builder Pattern.
   */
  protected ?string $label;

  /**
   * Component tree.
   *
   * @var ?array{'inputs': array, 'tree': array}
   */
  protected ?array $component_tree;

  /**
   * {@inheritdoc}
   */
  public function getComponentTree(): ComponentTreeItem {
    $component_tree_item = $this->createDanglingComponentTree();
    $component_tree_item->setValue($this->component_tree);
    return $component_tree_item;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    parent::calculateDependencies();

    $component_tree = $this->getComponentTree();
    $this->addDependencies($component_tree->get('tree')->getDependencies());

    // TRICKY: in theory, dependencies must also be calculated for the `inputs`
    // field prop. But, currently it can only contain StaticPropSources, and the
    // dependencies for those are tracked in the Component config entity.
    // @see \Drupal\experience_builder\Entity\Component::calculateDependencies()
    // @todo Revisit this when allowing more complex values in `inputs`, that are not dictated by/captured in the Component config entity.
    // @todo Revisit this in https://www.drupal.org/project/experience_builder/issues/3484666, where the above MIGHT change.

    return $this;
  }

  public static function preCreate(EntityStorageInterface $storage, array &$values) {
    if (!array_key_exists('id', $values)) {
      $values['id'] = self::generateId($values['label']);
    }
    parent::preCreate($storage, $values);
  }

  /**
   * Generates a valid ID from the given label.
   */
  private static function generateId(string $label): string {
    $id = mb_strtolower($label);

    $id = preg_replace('@[^a-z0-9_.]+@', '', $id);
    assert(is_string($id));
    // Furthermore remove any characters that are not alphanumerical from the
    // beginning and end of the transliterated string.
    $id = preg_replace('@^([^a-z0-9]+)|([^a-z0-9]+)$@', '', $id);
    assert(is_string($id));
    assert(is_string($id));
    if (strlen($id) > 23) {
      $id = substr($id, 0, 23);
    }
    assert(is_string($id));

    $query = \Drupal::entityTypeManager()->getStorage('pattern')->getQuery()->accessCheck(FALSE);
    $ids = $query->execute();
    $id_exists = in_array($id, $ids, TRUE);
    if ($id_exists) {
      $id = $id . '_' . (new Random())->machineName(8);
    }

    return $id;
  }

  /**
   * {@inheritdoc}
   *
   * This corresponds to `PatternPreview` in openapi.yml.
   *
   * @see docs/adr/0005-Keep-the-front-end-simple.md
   */
  public function normalizeForClientSide(): ClientSideRepresentation {
    $item = $this->getComponentTree();
    return ClientSideRepresentation::create(
      values: [
        'layoutModel' => ApiConfigControllers::convertComponentTreeItemToLayoutModel($item),
        'name' => $this->label(),
        'id' => $this->id(),
      ],
      preview: $item->toRenderable(),
    )->addCacheableDependency($this);
  }

  /**
   * {@inheritdoc}
   *
   * This corresponds to `Pattern` in openapi.yml.
   *
   * @see docs/adr/0005-Keep-the-front-end-simple.md
   */
  public static function createFromClientSide(array $data): static {
    $values = [];
    if (isset($data['id'])) {
      $values['id'] = $data['id'];
    }
    if (isset($data['name'])) {
      $values['label'] = $data['name'];
    }
    $entity = static::create($values);
    $entity->updateFromClientSide($data);
    return $entity;
  }

  /**
   * {@inheritdoc}
   *
   * This corresponds to `Pattern` in openapi.yml.
   *
   * @see docs/adr/0005-Keep-the-front-end-simple.md
   */
  public function updateFromClientSide(array $data): void {
    if (isset($data['layout']) && isset($data['model'])) {
      $this->set('component_tree', self::convertClientToServer($data['layout'], $data['model']));
    }

    foreach (array_diff_key($data, array_flip(['layout', 'model'])) as $key => $value) {
      if ($key == 'name') {
        $key = 'label';
      }
      $this->set($key, $value);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function refineListQuery(QueryInterface &$query, RefinableCacheableDependencyInterface $cacheability): void {
    // Nothing to do.
  }

}

==> experience_builder-0.4.1-alpha1.tar.gz: experience_builder/src/Entity/Pattern.php <==
<?php

declare(strict_types=1);

namespace Drupal\experience_builder\Entity;

use Drupal\Component\Utility\Random;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\experience_builder\ClientSideRepresentation;
use Drupal\experience_builder\Controller\ClientServerConversionTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\experience_builder\EntityHandlers\ContentCreatorVisibleXbConfigEntityAccessControlHandler;
use Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItemListInstantiatorTrait;
use Drupal\Core\Entity\Attribute\ConfigEntityType;
use Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItemList;

/**
 * @phpstan-import-type ComponentTreeItemArray from \Drupal\experience_builder\Plugin\Field\FieldType\ComponentTreeItemList
 */
#[ConfigEntityType(
  id: self::ENTITY_TYPE_ID,
  label: new TranslatableMarkup('Pattern'),
  label_singular: new TranslatableMarkup('pattern'),
  label_plural: new TranslatableMarkup('patterns'),
  label_collection: new TranslatableMarkup('Patterns'),
  admin_permission: self::ADMIN_PERMISSION,
  handlers: [
    'access' => ContentCreatorVisibleXbConfigEntityAccessControlHandler::class,
  ],
  entity_keys: [
    'id' => 'id',
    'label' => 'label',
    'status' => 'status',
  ],
  config_export: [
    'id',
    'label',
    'component_tree',
  ],
)]

final class Pattern extends ConfigEntityBase implements XbHttpApiEligibleConfigEntityInterface, ComponentTreeEntityInterface {

  public const string ENTITY_TYPE_ID = 'pattern';
  public const string ADMIN_PERMISSION = 'administer patterns';

  use ComponentTreeItemListInstantiatorTrait;
  use ClientServerConversionTrait;

  /**
   * Pattern entity ID.
   */
  protected string $id;

  /**
   * The human-readable label of the Experience Builder Pattern.
   */
  protected ?string $label;

  /**
   * Component tree.
   *
   * @var ?array<string, ComponentTreeItemArray>
   */
  protected ?array $component_tree;

  /**
   * {@inheritdoc}
   */
  public function getComponentTree(): ComponentTreeItemList {
    $component_tree = $this->createDanglingComponentTreeItemList();
    $component_tree->setValue(\array_values($this->component_tree ?? []));
    return $component_tree;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    parent::calculateDependencies();
    $this->addDependencies($this->getComponentTree()->calculateDependencies());
    return $this;
  }

  public static function preCreate(EntityStorageInterface $storage, array &$values) {
    if (!array_key_exists('id', $values)) {
      $values['id'] = self::generateId($values['label']);
    }
    parent::preCreate($storage, $values);
  }

  /**
   * Generates a valid ID from the given label.
   */
  private static function generateId(string $label): string {
    $id = mb_strtolower($label);

    $id = preg_replace('@[^a-z0-9_.]+@', '', $id);
    assert(is_string($id));
    // Furthermore remove any characters that are not alphanumerical from the
    // beginning and end of the transliterated string.
    $id = preg_replace('@^([^a-z0-9]+)|([^a-z0-9]+)$@', '', $id);
    assert(is_string($id));
    assert(is_string($id));
    if (strlen($id) > 23) {
      $id = substr($id, 0, 23);
    }
    assert(is_string($id));

    $query = \Drupal::entityTypeManager()->getStorage('pattern')->getQuery()->accessCheck(FALSE);
    $ids = $query->execute();
    $id_exists = in_array($id, $ids, TRUE);
    if ($id_exists) {
      $id = $id . '_' . (new Random())->machineName(8);
    }

    return $id;
  }

  /**
   * {@inheritdoc}
   *
   * This corresponds to `PatternPreview` in openapi.yml.
   *
   * @see docs/adr/0005-Keep-the-front-end-simple.md
   */
  public function normalizeForClientSide(): ClientSideRepresentation {
    $items = $this->getComponentTree();
    return ClientSideRepresentation::create(
      values: [
        'layoutModel' => $items->getClientSideRepresentation(),
        'name' => $this->label(),
        'id' => $this->id(),
      ],
      preview: $items->toRenderable($this, TRUE),
    )->addCacheableDependency($this);
  }

  /**
   * {@inheritdoc}
   *
   * This corresponds to `Pattern` in openapi.yml.
   *
   * @see docs/adr/0005-Keep-the-front-end-simple.md
   */
  public static function createFromClientSide(array $data): static {
    $values = [];
    if (isset($data['id'])) {
      $values['id'] = $data['id'];
    }
    if (isset($data['name'])) {
      $values['label'] = $data['name'];
    }
    $entity = static::create($values);
    $entity->updateFromClientSide($data);
    return $entity;
  }

  /**
   * {@inheritdoc}
   *
   * This corresponds to `Pattern` in openapi.yml.
   *
   * @see docs/adr/0005-Keep-the-front-end-simple.md
   */
  public function updateFromClientSide(array $data): void {
    if (isset($data['layout']) && isset($data['model'])) {
      $this->set('component_tree', self::convertClientToServer($data['layout'], $data['model']));
    }

    foreach (array_diff_key($data, array_flip(['layout', 'model'])) as $key => $value) {
      if ($key == 'name') {
        $key = 'label';
      }
      $this->set($key, $value);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function refineListQuery(QueryInterface &$query, RefinableCacheableDependencyInterface $cacheability): void {
    // Nothing to do.
  }

  /**
   * {@inheritdoc}
   */
  public function set($property_name, $value): self {
    if ($property_name === 'component_tree') {
      // Ensure predictable order of tree items.
      $value = self::generateComponentTreeKeys($value);
    }
    return parent::set($property_name, $value);
  }

}
